/*------------------------------------------------------
 * urc.c: H8 based USB motor controller driver
 * $Id: urc.c,v 0.2 2002/11/20
 * Copyright (c) 2002 iXs resarch corp.
 * based on
 * Copyright (c) 2001 Takashi Minato
 * Copyright (c) 2001-2002 Noriaki Mitsunaga
 *------------------------------------------------------
 */
/* ºݤ˻Ȥž⡼ɤϥХ륯žΤ
   Ȥͽ */
#include <linux/devfs_fs_kernel.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/time.h>
#include <linux/usb.h>

#include "urc.h"
#include "urcc.h"

/*------------------------------*/
/* Ƽ                     */
/*------------------------------*/
#define URC_MINOR_BASE    140    /* ޥֹ */
#define MAX_DEVICES       32     /* We can have up to this number of 
				    device plugged in at once */

#define EP_SCMD    2 /* Endpoint to write gains and desired values */
#define EP_READ    5 /* Endpoint to read current status */

#define DRIVER_VERSION "v0.2"
#define DRIVER_AUTHOR "Noriaki Mitsunaga"
#define DRIVER_DESC   "H8 based USB motor controller driver"

/* Supported devices */
struct usb_devices {
  __u16   idVendor;
  __u16   idProduct;
};
static struct usb_devices supported_devices[] = {
  {0xff8, 0x0010},     /* iXs Research version */
  {0, 0}
};

/* the global usb devfs handle */
extern devfs_handle_t usb_devfs_handle;

/* ǡ¸¤ */
typedef struct {
    struct usb_device   *udev;       /* USB ǥХǡ¸  */
    devfs_handle_t	devfs;	     /* devfs device node */
    unsigned char       minor;	     /* the minor number for this device */
    struct urb          *wurb;
    struct urb          *rurb;
    struct urb          *irq;         /* 󥿥ץ URB      */
    wait_queue_head_t   rwait;
    wait_queue_head_t   wwait;
    wait_queue_head_t   wait;        /* 󥿥ץԤ塼  */
    wait_queue_head_t   remove_ok;   /* DISCONNECTνλԤ塼 */
    char                data[2];     /* 󥿥ץȼХåե */
    struct urcin          rbuf_;       /* read buffer */
#if 1
	char		dmy[50];	/* I'm not sure but after usb_submit()
					   memory region around here is broken
					   by USB upper part */
#endif
    struct urcin          rbuf;        /* read buffer */
    unsigned char       wbuf[64];    /* write buffer */
    int                 isopen;      /* ¿ťץػߥե饰   */
    int                 buffering;
    struct semaphore	sem;	     /* lock semaphore */
    struct semaphore	wsem;	     /* lock semaphore */
    struct semaphore	rsem;	     /* lock semaphore */
} urc_data_t;

struct urc {
    urc_data_t *dev;
    int                 ep_in;
    int                 ep_out;
    int                 last;
    int                 (*write)(urc_data_t*, int, const char *, size_t);
    int                 (*read)(struct urc*, char *, size_t);
};

/* ե뤴Ȥδΰ */
typedef struct{
  unsigned long f_version;   //  f_version
  wait_queue_head_t wait;    // ̲Եΰ
  int sleep_mode;            // 0: Ƥ 1: Ԥ
  unsigned long long wtime;  // Ǹ˵
  unsigned long wakeup_jiffies;  // ٤
  unsigned int period;       // 
}FileInfo;

static __devinitdata struct usb_device_id device_table [] = {
	{ USB_DEVICE(0x0ff8, 0x0010) },     /* iXs Research version */
	{ } /* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, device_table);

/* Declarations */
static int urc_open(struct inode *inode, struct file *file);
static int urc_close(struct inode *inode, struct file *file);
static void* urc_probe(struct usb_device *dev, unsigned int ifnum, const struct usb_device_id *id);
static void urc_disconnect(struct usb_device *dev, void *ptr);
static int urc_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static ssize_t urc_write(struct file *file, const char *buf, size_t count, loff_t *ppos);
static ssize_t urc_read(struct file *file, char *buf, size_t count, loff_t *ppos);
#if 0
static unsigned int urc_poll(struct file *file, poll_table *ptab);
#endif
static int __init urc_init(void);
static void __exit urc_exit(void);

/* array of pointers to our devices that are currently connected */
static urc_data_t          *minor_table[MAX_DEVICES];

/* lock to protect the minor_table structure */
static DECLARE_MUTEX (minor_table_mutex);

static struct file_operations urc_fileopes = {
    owner:          THIS_MODULE,
    read:           urc_read,
    write:          urc_write,
#if 0
    poll:           urc_poll,
#endif
    ioctl:          urc_ioctl,
    open:           urc_open,
    release:        urc_close,
};

/* Functions */

static void read_bulk_callback( struct urb *urb )
{
	urc_data_t* dev = urb->context;
	int res;

	if ( !dev || dev->isopen == 0) 
		return;

	if ( urb->status != 0) {
		info("urc: read callback status %d", urb->status);
	}

	if (down_trylock (&dev->rsem) == 0) {
		memcpy(&dev->rbuf, &dev->rbuf_, sizeof(dev->rbuf));
		up (&dev->rsem);
	}

	if (dev->buffering == 0)
	  return;

	FILL_BULK_URB( dev->rurb, dev->udev,
			usb_rcvbulkpipe(dev->udev, EP_READ),
			&dev->rbuf_, sizeof(dev->rbuf_), 
			read_bulk_callback, dev );
	if ( (res = usb_submit_urb(dev->rurb)) )
		warn( __FUNCTION__ " failed submitting rurb %d", res);

	/* Wake up if there is waiting queue */
        wake_up_interruptible(&dev->rwait);
}

static inline int read_wait(struct urc* urc, char *data, size_t count)
{   /* Wait for next packet comes */
    urc_data_t *dev = urc->dev;
    int ep = urc->ep_in;
    int r, rlen;
    char *rbuf;

    if ((rbuf = kmalloc(count, GFP_KERNEL)) == NULL) {
      printk(KERN_INFO "No memory\n");
      return -1;
    }

    r = usb_bulk_msg(dev->udev,
    			usb_rcvbulkpipe(dev->udev, ep),
			rbuf, count, &rlen, HZ*5);
    if (r < 0 || rlen != count) {
	printk(KERN_INFO "error\n");
	kfree(rbuf);
        return -1;
    }

    if (copy_to_user(data, rbuf, count) != 0) {
      kfree(rbuf);
      return -EFAULT;
    }

    kfree(rbuf);
    return count;
}

static inline int read_buf(struct urc* urc, char *data, size_t count)
{   /* read from buffer, ep is ignored */
  urc_data_t *dev = urc->dev;

  down (&dev->rsem);
  if (urc->last == dev->rbuf.time) {
    /* ɤtime stamp Ȱʤ鼡Ԥ */
    up (&dev->rsem);
    interruptible_sleep_on(&dev->rwait);
    if (signal_pending(current)) {
        /* ʥ */
        return  -EINTR;
    }
    if (dev->udev == NULL) {
        /* ˥ǥХȴ줿 */
        return -ENODEV;
    }
    down (&dev->rsem);
  }
  urc->last = dev->rbuf.time;
  if (copy_to_user(data, &dev->rbuf, count) != 0) {
    up (&dev->rsem);
    return -EFAULT;
  }
  up (&dev->rsem);
  return count;
}

static ssize_t urc_read(struct file *file, char *buf, 
                          size_t count, loff_t *ppos)
{
  struct urc* urc = (struct urc*)file->private_data;
  urc_data_t* dev = urc->dev;

  dbg(__FUNCTION__ " - minor %d, count = %d", dev->minor, count);

  if(count != 64) return -1;

  /* lock this object */
  down (&dev->sem);

  /* verify that the device wasn't unplugged */
  if (dev->udev == NULL) {
	up (&dev->sem);
      	return -ENODEV;
  }

  if (urc->read(urc, buf, count) < 0){
    printk(KERN_INFO "read failed\n");
    up (&dev->sem);
    return -1;
  }

  /* unlock the device */
  up (&dev->sem);

  return count;
}

static int write_wait(urc_data_t *dev, int ep, const char *buf, size_t count)
{      /* send and wait completion */
       int r, rlen;
       unsigned char *wbuf;
       
       if ((wbuf = kmalloc(count, GFP_KERNEL)) == NULL) {
	 printk(KERN_INFO "No memory\n");
	 return -1;
       }

       copy_from_user(wbuf, buf, count);
       r = usb_bulk_msg(dev->udev, 
    			usb_sndbulkpipe(dev->udev, ep),
			wbuf, count, &rlen, HZ*5);
       if (r < 0 || rlen != count) {
	 printk(KERN_INFO "Error sending data\n");
       }
       kfree(wbuf);
       return rlen;
}

static void write_bulk_callback( struct urb *urb )
{
	urc_data_t *dev = urb->context;

	if ( !dev || dev->udev == NULL )
		return;
		
	if ( urb->status )
		info("urc: write callback status %d", urb->status);

	/* Wake up if there is waiting queue */
        wake_up_interruptible(&dev->wwait);
}

static int write_submit(urc_data_t *dev, int ep, const char *buf, size_t count)
{      /* submit for sending and do not wait completion */
        int r;
	unsigned long flags;

	spin_lock_irqsave(&dev->wurb->lock, flags);

        if (dev->wurb->status == -EINPROGRESS || dev->wurb->status == -ECONNRESET ||
            dev->wurb->status == -ECONNABORTED) {
                spin_unlock_irqrestore(&dev->wurb->lock, flags);
    		interruptible_sleep_on(&dev->wwait);
    		if (signal_pending(current)) {
    		    /* ʥ */
    		    return  -EINTR;
    		}
    		if (dev->udev == NULL) {
    		    /* ˥ǥХȴ줿 */
    		    return -ENODEV;
    		}
        } else {
                spin_unlock_irqrestore(&dev->wurb->lock, flags);
	}

       copy_from_user(dev->wbuf, buf, sizeof(dev->wbuf));
       FILL_BULK_URB( dev->wurb, dev->udev,
		     usb_sndbulkpipe(dev->udev, ep),
		     dev->wbuf, sizeof(dev->wbuf), 
		     write_bulk_callback, dev);
       if ( (r = usb_submit_urb(dev->wurb)) != 0 ) {
	 warn( __FUNCTION__ " failed submitting wurb %d", r);
	 return -1;
       }

       return count;
}

static ssize_t urc_write(struct file *file, const char *buf, 
                           size_t count, loff_t *ppos)
{
     struct urc* urc = (struct urc*)file->private_data;
     urc_data_t* dev = urc->dev;
     int r;
 
     down (&dev->sem);
     if (dev->udev == NULL) {
       up (&dev->sem);
       return -1;
     }

     if(count != 64) return -1;

     r = urc->write(dev, urc->ep_out, buf, count);

     up (&dev->sem);
     return r;
}

static int urc_ioctl(struct inode *inode, struct file *file,
                       unsigned int cmd, unsigned long arg)
{
    struct urc* urc = (struct urc*)file->private_data;
    urc_data_t* dev = urc->dev;
    struct usb_device *udev = dev->udev;

    switch(cmd) {
    case URC_GET_VENDOR:
      if(copy_to_user((int*)arg, &udev->descriptor.idVendor, sizeof(int)))
	return -EFAULT;
      break;
    case URC_GET_PRODUCT:
      if(copy_to_user((int*)arg, &udev->descriptor.idProduct, sizeof(int)))
	return -EFAULT;
      break;
    case URC_READ: /* end point 5 */
      urc->ep_in = EP_READ;
      break;
    case URC_SET:     /* end point 2 */
      urc->ep_out = EP_SCMD;
      urc->write = write_submit;
      break;
    default:
      return -ENOIOCTLCMD;
    }
    //      printk("URC: IN %d OUT %d\n", urc->ep_in, urc->ep_out);

    return 0;
}

static int urc_open(struct inode *inode, struct file *file)
{
  urc_data_t* dev = NULL;
  struct urc* urc;
  int subminor;
  int retval = 0;

  dbg(__FUNCTION__);

  subminor = MINOR (inode->i_rdev) - URC_MINOR_BASE;
  if ((subminor < 0) ||
      (subminor >= MAX_DEVICES)) {
    return -ENODEV;
  }

  /* increment our usage count for the module */
  MOD_INC_USE_COUNT;

  /* lock our minor table and get our local data for this minor */
  down (&minor_table_mutex);
  dev = minor_table[subminor];
  if (dev == NULL) {
    up (&minor_table_mutex);
    MOD_DEC_USE_COUNT;
    return -ENODEV;
  }

  /* lock this device */
  down (&dev->sem);
  
  /* unlock the minor table */
  up (&minor_table_mutex);

  init_waitqueue_head(&dev->rwait);
  init_waitqueue_head(&dev->wwait);
#if 0
  /* 󥿥ץ wait 塼 */
  init_waitqueue_head(&dev->wait);
  /* DISCONNECT νλԤ塼 */
  init_waitqueue_head(&dev->remove_ok);
#endif
  /* Save our object in the file's private structure */
  if ((urc = kmalloc(sizeof(urc), GFP_KERNEL)) == NULL) {
    printk(KERN_INFO "no memory\n");
    up (&dev->sem);
    return -ENODEV;
  }
  urc->dev = dev;
  urc->read = read_wait;
  urc->write = write_wait;
  urc->ep_in = EP_READ;
  urc->ep_out = EP_SCMD;
  urc->last = 0;
  dev->isopen ++;
  file->private_data = urc;
#if 0
  /* 󥿥ץ URB դ */
  usb_submit_urb(&dev->irq);
#endif
  
  /* unlock this device */
  up (&dev->sem);

  return retval;
}

static int urc_close(struct inode *inode, struct file *file)
{
    struct urc* urc = (struct urc*)file->private_data;
    urc_data_t* dev = urc->dev;

    down (&dev->sem);
    if (dev->isopen > 0) {
        dev->isopen --;
	if (dev->buffering > 0 && dev->isopen == 0)
		dev->buffering = 0;
#if 0
        wake_up(&dev->remove_ok);
        /* 󥿥ץ URB ʤ */
        usb_unlink_urb(&dev->irq);
#endif
        usb_unlink_urb(dev->wurb);
        usb_unlink_urb(dev->rurb);
	MOD_DEC_USE_COUNT;
    }
    up (&dev->sem);

    kfree(urc);

    return 0;
}

#if 0
static unsigned int urc_poll(struct file *file, 
			       poll_table *ptab)
{
  FileInfo *fip=(FileInfo *)(file->private_data);
  poll_wait(file,&(fip->wait),ptab);
  fip->sleep_mode=1;
  if(fip->wakeup_jiffies <= jiffies){
    fip->sleep_mode=0;
    fip->wakeup_jiffies=    // ˵
      jiffies+(fip->period)-(jiffies-fip->wakeup_jiffies)%(fip->period);
    return POLLIN|POLLRDNORM;                        // ɤ߹OK
  }else
    return 0;
}
#endif
#if 0
/*-----------------------------------*/
/* 󥿥ץȥǡԤ        */
/*-----------------------------------*/
static int urc_intr_wait(urc_data_t* dev)
{
    int ret = 0;
    /* 󥿥ץȼԤ */
    interruptible_sleep_on(&dev->wait); /* WAIT  usbkit_irq ؿ */
    if (signal_pending(current)) {
        /* ʥ */
        ret =  -EINTR;
    }
    if (dev->udev == NULL) {
        /* ˥ǥХȴ줿 */
        return -ENODEV;
    }
    return ret;
}

/*-----------------------------------*/
/* USB 󥿥ץȥϥɥ          */
/*-----------------------------------*/
static void urc_irq(struct urb *urb)
{
    urc_data_t* dev;
    if (urb->status) {
        return;
    }
    dev = urb->context;
    if (dev->data[0]) {
        /* Хܤʳʤгȯ */
        wake_up_interruptible(&dev->wait);
    }
}
#endif
/*==========================================================*/
/*  USB ǥХɥ饤Ͽ/                           */
/*==========================================================*/
static void* urc_probe(struct usb_device *udev, unsigned int ifnum,
			 const struct usb_device_id *id)
{
	struct usb_interface_descriptor *inf;
	struct usb_interface *interface;
	struct usb_config_descriptor    *config;
	urc_data_t *dev;
	int endpoint_flag = 0;
	int minor, n, intr_interval, i;

	/* Is this the USB robot controller ? */
	for (i=0; supported_devices[i].idVendor!=0; i++) {
	  if ((udev->descriptor.idVendor  == supported_devices[i].idVendor) &&
	      (udev->descriptor.idProduct == supported_devices[i].idProduct))
	    break;
	}
	if (supported_devices[i].idVendor == 0)
	  return NULL;

	/* select a "subminor" number (part of a minor number) */
	down (&minor_table_mutex);
	for (minor = 0; minor < MAX_DEVICES; ++minor) {
	  if (minor_table[minor] == NULL)
	    break;
	}
	if (minor >= MAX_DEVICES) {
	  info ("Too many devices plugged in, can not handle this device.");
	  return NULL;
	}
	
	dev = kmalloc(sizeof(urc_data_t), GFP_KERNEL);
	if(dev == NULL){
	  printk(KERN_INFO "no memory\n");
	  return NULL;
	}
	config = udev->actconfig;
	interface = &udev->actconfig->interface[ifnum];
	inf    = &interface->altsetting[0];
	
	minor_table[minor] = dev;
	init_MUTEX (&dev->sem);
	init_MUTEX (&dev->rsem);
	init_MUTEX (&dev->wsem);
	dev->udev    = udev;
	/* dev->interface = interface; */
	dev->minor = minor;

	/* ɥݥȲ(ɥݥ1,2,5,6å) */
	intr_interval = 0;
	
	for(n = 0; n < inf->bNumEndpoints; n++){
#if 0
	  printk("%d %d %d %d\n", n, 
		 inf->endpoint[n].bEndpointAddress&USB_ENDPOINT_NUMBER_MASK,
		 inf->endpoint[n].bEndpointAddress&USB_ENDPOINT_DIR_MASK,
		 inf->endpoint[n].bmAttributes&USB_ENDPOINT_XFERTYPE_MASK);
#endif
	  switch(inf->endpoint[n].bEndpointAddress&USB_ENDPOINT_NUMBER_MASK){
	  case 1: /* Х륯 IN */
	    /* DIR bit0ʤۥȤǥХ(OUT)
	       1ʤǥХۥȤ(IN) */
	    if(!inf->endpoint[n].bEndpointAddress&USB_ENDPOINT_DIR_MASK)
	      break;
	    if((inf->endpoint[n].bmAttributes&USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK){
	      endpoint_flag ++;
	    }
	    break;
	  case 2: /* Х륯 OUT */
	    /* DIR bit0ʤۥȤǥХ(OUT)
	       1ʤǥХۥȤ(IN) */
	    if(inf->endpoint[n].bEndpointAddress&USB_ENDPOINT_DIR_MASK)
	      break;
	    if((inf->endpoint[n].bmAttributes&USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK){
	      endpoint_flag ++;
	    }
	    break;
	  case 5: /* Х륯 IN */
	    /* DIR bit0ʤۥȤǥХ(OUT)
	       1ʤǥХۥȤ(IN) */
	    if(!inf->endpoint[n].bEndpointAddress&USB_ENDPOINT_DIR_MASK)
	      break;
	    if((inf->endpoint[n].bmAttributes&USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK){
	      endpoint_flag ++;
	    }
	    break;
	  case 6: /* Х륯 OUT */
	    /* DIR bit0ʤۥȤǥХ(OUT)
	       1ʤǥХۥȤ(IN) */
	    if(inf->endpoint[n].bEndpointAddress&USB_ENDPOINT_DIR_MASK)
	      break;
	    if((inf->endpoint[n].bmAttributes&USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK){
	      endpoint_flag ++;
	    }
	    break;
	  }
	}
	if(endpoint_flag != 4){
	  printk("USBKIT: endpoint number missmatch!!\n");
	  kfree(dev);
	  return NULL;
	}
	dev->rurb = usb_alloc_urb(0);
	if (dev->rurb == NULL) {
		printk(KERN_INFO "usb_alloc_urb() failed.\n");
		kfree(dev);
		return NULL;
	}
	dev->wurb = usb_alloc_urb(0);
	if (dev->wurb == NULL) {
		printk(KERN_INFO "usb_alloc_urb() failed.\n");
		kfree(dev);
		return NULL;
	}
	spin_lock_init(&dev->wurb->lock);
	dev->isopen = 0;
	dev->buffering = 0;

	/* let the user know what node this device is now attached to */
	info ("USB robot controller (%x/%x/%x) now attached to urc%d", 
	      udev->descriptor.idVendor, 
	      udev->descriptor.idProduct, 
	      udev->descriptor.bcdDevice, 
	      dev->minor);
    
	up (&minor_table_mutex);
	return dev;
}

static void urc_disconnect(struct usb_device *udev, void *ptr)
{
        urc_data_t* dev = ptr;
	int minor = dev->minor;

	down (&minor_table_mutex);
	down (&dev->sem);
	down (&dev->wsem);
	down (&dev->rsem);

#if 0
	/* remove our devfs node */
	devfs_unregister(dev->devfs);
#endif

	if (dev->isopen > 0) {
	  /* ǥХػߤ뤿 struct usb_driver ѿ򥯥ꥢ */
	  dev->udev = NULL;
	  up (&dev->sem);
	  up (&dev->wsem);
	  up (&dev->rsem);

	  /* ԤĤʤ褦˲ */
	  init_waitqueue_head(&dev->rwait);
	  init_waitqueue_head(&dev->wwait);
#if 0
	  wake_up_interruptible(&dev->wait);
	  /* ɥ饤Х꥽ǽˤʤޤԤ */
	  sleep_on(&dev->remove_ok);
#endif
	}
	/* ɥ饤Х꥽ */
	minor_table[minor] = NULL;
	usb_free_urb(dev->rurb);
	usb_free_urb(dev->wurb);
	kfree(dev);
	info("urc%d now disconnected", minor);
	up (&minor_table_mutex);
}

/*==========================================================*/
/*  ⥸塼Ͽ                                    */
/*==========================================================*/
/*---------------------------------*/
/* USB ɥ饤Ͽѥǡ        */
/*---------------------------------*/
static struct usb_driver urc_driver = {
    name:		"urc",           /* ̾            */
    id_table:		device_table,
    probe:		urc_probe,       /* PROBE ؿ      */
    disconnect:		urc_disconnect,  /* DISCONNECT ؿ */
    fops:		&urc_fileopes,   /* եɥ饤дؿ    */
    minor:		URC_MINOR_BASE,  /* Υɥ饤ФΥޥʡֹ */
};

/*-------------------------------------------*/
/* ⥸塼Ͽؿ                        */
/*-------------------------------------------*/
static int __init urc_init(void)
{
	int result;
	/* USB ǥХкؿϿ */
	if ((result = usb_register(&urc_driver)) < 0) {
		err("usb_register failed for the "__FILE__
		    " driver. Error number %d", result);
		return -1;
	}

	info(DRIVER_DESC " " DRIVER_VERSION);

	return 0;
}

/*--------------------------------------------*/
/* ⥸塼ؿ                         */
/*--------------------------------------------*/
static void __exit urc_exit(void)
{
    /* USB ǥХкؿκ */
    usb_deregister(&urc_driver);
}

module_init(urc_init);
module_exit(urc_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
